/**
*
* $Id: b58.cpp
* (c) 2005 Sputnik -- erik_h_anderson@hotmail.com
* Based on CVEL Centaur-G code
* Ervins Reinverts -- ervins@gmail.com
* Based on CVEL Fregat code
* (c) 2004 David Henderson -- orbiter@aibs.org.uk
* Licence LGPL -- http://www.gnu.org/licenses/lgpl.txt
*
* Autopilot by rjcroy
*
* You may use this source code in your addon provided that if you
* publish your addon either privately or publicly the end user
* recieves a copy of the source code including any modifications
* you may have made. 
*
*/

#include "b58.h"


// 1. vertical lift component (wings and body)
// Borrowed from Mike Norman's Shuttle code

void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*15.0;
	static const double istep = 1.0/step;
	static const int nabsc = 25;
	//                              -180 -165  -150 -135  -120 -105 -90 -75.0  -60.0  -45.0  -30.0  -15.0   0.0   15.0      30.0     45.0  60.0  75.0  90.0 105.0  120.0  135.0  150.0  165.0 180.0  
	static const double CL[nabsc] = {0.1, 0.17, 0.2, 0.2, 0.17, 0.1, 0, -0.11, -0.24, -0.38,  -0.5,  -0.5,  0.02, 0.655,     0.46,   0.32, 0.12, 0.05, 0.0, -0.16, -0.26, -0.29, -0.24, -0.1, 0.1};
	static const double CM[nabsc] = {  0,    0,   0,   0,    0,   0, 0,     0,     0,  0.002,  0.004, 0.0025,0.0012,      0,-0.0012,-0.0007,    0,    0,   0,     0,     0,     0,     0,    0,   0};
	// lift and moment coefficients from -180 to 180 in 15 degree steps.
	// This uses a documented lift slope of 0.0437/deg, everything else is rather ad-hoc

	// change for drag increase at high Mach:
	// Subsonic Equivalents

	double Cl0, Cl1, Cd0, Cd1, Cd2;
	if(M>=5.0) {
		Cl0 = -0.187040;
		Cl1 =  1.275600;
		Cd0 =  0.178540;
		Cd1 = -0.402900;
		Cd2 =  1.539960;
		*cl = Cl0 + Cl1*aoa;
		*cd = Cd0 + Cd1*aoa + Cd2*aoa* aoa;
	}
 	aoa += PI;
	int idx = max (0, min (23, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cm = (CM[idx] + (CM[idx+1]-CM[idx])*d);
	if(M<1.25) { 
 		*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
 		*cd = 0.03 + 0.2*oapiGetInducedDrag (*cl, 6.5, 0.92);
	}
	else if(M<5.0) { 
 		*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
 		*cd = 0.035 + 0.25*oapiGetInducedDrag (*cl, 4.1, 0.85);
// 		*cd = 0.06 + oapiGetInducedDrag (*cl, 2.266, 0.6);
	}
}


// 2. horizontal lift component (vertical stabilisers and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {       0,    +0.3,      0,   -0.3,  +0.3,     0,   -0.3,      0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.015 + oapiGetInducedDrag (*cl, 2.5, 0.8) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	
	if (M > 1.0) *cd *= 0.6;
	else *cd *= 0.3;
}


B58::B58 (OBJHANDLE hObj, int fmodel)
: COMMON_VESSEL_EX (hObj)
{
	strcat (meshname,"b58");

	hB58				= oapiLoadMeshGlobal (meshname);
	hWeapon				= oapiLoadMeshGlobal ("BLU2p2");
	hTank				= oapiLoadMeshGlobal ("BLU2p1");

	tex = oapiRegisterExhaustTexture("RealExhaust");

	/* set the propellant handles to NULL so I can check if I need to 
	/* create them in the SetMainStage() function. */

	ph_b58=NULL;
	ph_tank=NULL;
	bBoosterMade=false;
	bControlsMade=false;
	gear_status=DOOR_CLOSED;
	gear_proc=0.0;
	oldsimt = -0.01;
	Vold = _V(0,0,0);
	stage=2;
	BrakesOn = false;
	
	hB58				= oapiLoadMeshGlobal (meshname);

	MyID=ConnectToOrbiterSoundDLL3(GetHandle());
	RequestLoadVesselWave3(MyID,PLAYGEAR,"Sound\\b58\\gear.wav",BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave3(MyID,PLAYBANG,"Sound\\b58\\stg2sep.wav",BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave3(MyID,PLAYJETT,"Sound\\b58\\fasep.wav",BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave3(MyID,PLAYJET,"Sound\\b58\\jet.wav",BOTHVIEW_FADED_FAR);

	/* CVE-lite function ! */
	/* IMPORTANT */
	/* This function loops through the pending GDPS requests. Have it run
	/* in the constructor so it can set up things like the InitFuelLevel
	/* (initial fuel level) variable, and load the payloads */
	CheckCentralMessage();   
 
	MakeAnimations();

	Firstsweep = true;

}


B58::~B58()
{

}

void B58::MakeAnimations()
{

	anim_gear = 0.0;

	anim_gear = CreateAnimation (1);
	anim_lelevator = CreateAnimation (0.5);
	anim_relevator = CreateAnimation (0.5);
	anim_laileron = CreateAnimation (0.5);
	anim_raileron = CreateAnimation (0.5);
	anim_rudder = CreateAnimation (0.5);
	anim_lelevon = CreateAnimation (0.5);
	anim_relevon = CreateAnimation (0.5);


/*	static UINT LGearDoorGrp[1] = {10};
	static MGROUP_ROTATE LGearDoor (0, LGearDoorGrp, 1,
		_V(-1.038, 0.68,0.3), _V(0,0,1), (float)(-130*RAD));
*/
/*	static UINT LFSuptGrp[1] = {7};
	static MGROUP_ROTATE LFSupt (0, LFSuptGrp, 1,
		_V(-1.3,-0.7,-5.17), _V( 1,0,-0.45), (float)( 110*RAD));
	static UINT LFGrGrp[2] = {1,6};
	static MGROUP_ROTATE LFGear (0, LFGrGrp, 2,
		_V(-1.3,-0.11,-1.85), _V( 1,0,-0.45), (float)(-90*RAD));

	static UINT LASuptGrp[1] = {8};
	static MGROUP_ROTATE LASupt (0, LASuptGrp, 1,
		_V(-1.3,-0.9,-14.48), _V( 1,0,-0.45), (float)( 110*RAD));
	static UINT LAGrGrp[2] = {2,9};
	static MGROUP_ROTATE LAGear (0, LAGrGrp, 2,
		_V(-1.3,-0.285,-11.19), _V( 1,0,-0.45), (float)(-90*RAD));

	static UINT RFSuptGrp[1] = {11};
	static MGROUP_ROTATE RFSupt (0, RFSuptGrp, 1,
		_V( 1.3,-0.7,-5.17), _V( 1,0,0.45), (float)( 110*RAD));
	static UINT RFGrGrp[2] = {3,10};
	static MGROUP_ROTATE RFGear (0, RFGrGrp, 2,
		_V( 1.3,-0.11,-1.85), _V( 1,0, 0.45), (float)(-90*RAD));

	static UINT RASuptGrp[1] = {12};
	static MGROUP_ROTATE RASupt (0, RASuptGrp, 1,
		_V( 1.3,-0.9,-14.48), _V( 1,0,0.45), (float)( 110*RAD));
	static UINT RAGrGrp[2] = {4,13};
	static MGROUP_ROTATE RAGear (0, RAGrGrp, 2,
		_V( 1.3,-0.285,-11.19), _V( 1,0, 0.45), (float)(-90*RAD));
*/	
	static UINT NoseGrStrutGrp[2] = {6,7};
	static MGROUP_ROTATE NoseGearStrut (0, NoseGrStrutGrp, 2,
		_V( 0.0, -1.585, 8.940), _V( 1,0, 0), (float)( 100*RAD));
	static UINT NoseGrYokeGrp[1] = {5};
	static MGROUP_ROTATE NoseGearYoke (0, NoseGrYokeGrp, 1,
		_V( 0.0, -0.76, 10.350), _V( 1,0, 0), (float)(-80*RAD));
	static UINT NoseGrArmGrp[1] = {4};
	static MGROUP_ROTATE NoseGearArm (0, NoseGrArmGrp, 1,
		_V( 0.0, -0.606, 9.976), _V( 1,0, 0), (float)(-65*RAD));
	static UINT LNoseGrDrGrp[1] = {9};
	static MGROUP_ROTATE LNoseGrDr (0, LNoseGrDrGrp, 1,
		_V(-0.458, -0.836, 8.761), _V( 0,0, 1), (float)(-100*RAD));
	static UINT RNoseGrDrGrp[1] = {8};
	static MGROUP_ROTATE RNoseGrDr (0, RNoseGrDrGrp, 1,
		_V( 0.458, -0.836, 8.761), _V( 0,0, 1), (float)( 100*RAD));

	
	static UINT RInrGrDrGrp[1] = {27};
	static MGROUP_ROTATE RInrGrDr (0, RInrGrDrGrp, 1,
		_V( 1.651, -0.500, 2.261), _V( 0,0, 1), (float)(-100*RAD));
	static UINT ROutrGrDrGrp[1] = {28};
	static MGROUP_ROTATE ROutrGrDr (0, ROutrGrDrGrp, 1,
		_V( 2.596, -0.500, 2.261), _V( 0,0, 1), (float)( 100*RAD));
	static UINT RLongStrutGrp[1] = {15};
	static MGROUP_ROTATE RLongStrut (0, RLongStrutGrp, 1,
		_V( 2.142, -0.375, 0.342), _V( 1,0, 0), (float)(-35*RAD));
	static UINT RAFrameGrp[1] = {14};
	static MGROUP_ROTATE RAFrame (0, RAFrameGrp, 1,
		_V( 2.142, -0.375,-2.573), _V( 1,0, 0), (float)( 75*RAD));
	static UINT RStrutGrp[1] = {13};
	static MGROUP_ROTATE RStrut (0, RStrutGrp, 1,
		_V( 2.142, -1.175,-1.393), _V( 1,0, 0), (float)(-90*RAD));
	static UINT RWheelGrp[3] = {12, 11, 10};
	static MGROUP_ROTATE RWheel (0, RWheelGrp, 3,
		_V( 2.142, -1.312,-2.375), _V( 1,0, 0), (float)(-180*RAD));

	static UINT LInrGrDrGrp[1] = {29};
	static MGROUP_ROTATE LInrGrDr (0, LInrGrDrGrp, 1,
		_V(-1.651, -0.500, 2.261), _V( 0,0, 1), (float)( 100*RAD));
	static UINT LOutrGrDrGrp[1] = {30};
	static MGROUP_ROTATE LOutrGrDr (0, LOutrGrDrGrp, 1,
		_V(-2.596, -0.500, 2.261), _V( 0,0, 1), (float)(-100*RAD));
	static UINT LLongStrutGrp[1] = {21};
	static MGROUP_ROTATE LLongStrut (0, LLongStrutGrp, 1,
		_V(-2.142, -0.375, 0.342), _V( 1,0, 0), (float)(-35*RAD));
	static UINT LAFrameGrp[1] = {20};
	static MGROUP_ROTATE LAFrame (0, LAFrameGrp, 1,
		_V(-2.142, -0.375,-2.573), _V( 1,0, 0), (float)( 75*RAD));
	static UINT LStrutGrp[1] = {19};
	static MGROUP_ROTATE LStrut (0, LStrutGrp, 1,
		_V(-2.142, -1.175,-1.393), _V( 1,0, 0), (float)(-90*RAD));
	static UINT LWheelGrp[3] = {16, 17, 18};
	static MGROUP_ROTATE LWheel (0, LWheelGrp, 3,
		_V(-2.142, -1.312,-2.375), _V( 1,0, 0), (float)(-180*RAD));


/*
	static UINT LTailGearGrp[1] = {23};
	static MGROUP_ROTATE LTailGear (0, LTailGearGrp, 1,
		_V( 0.74,-0.35,-4.0), _V(0,0.0,1.0), (float)(-45*RAD));
	static UINT LTailTransGrp[1] = {23};
	static MGROUP_TRANSLATE LTailTrans (0, LTailTransGrp, 1,
		_V( 0.0,0.0,-0.30));
	static UINT RTailGearGrp[1] = {22};
	static MGROUP_ROTATE RTailGear (0, RTailGearGrp, 1,
		_V(-0.74,-0.35,-4.0), _V(0,0.0,1.0), (float)( 45*RAD));
	static UINT RTailTransGrp[1] = {22};
	static MGROUP_TRANSLATE RTailTrans (0, RTailTransGrp, 1,
		_V( 0.0,0.0,-0.30));
*/
// Known Groups:

// 0 Body
// 1 L Aileron
// 2 R Aileron
// 3 Rudder
// 4 NoseArm
// 5 NoseYoke
// 6 NWheels
// 7 NStrut
// 8 RNGDr
// 9 LNGDr
// 10 RFWheels
// 11 RAWheels
// 12 RTree
// 13 RStrut
// 14 RAFrame
// 15 RLongstrut
// 16 LFWheels
// 17 LAWheels
// 18 LTree
// 19 LStrut
// 20 LAFrame
// 21 LLongStrut
// 22 Aft Hull
// 23 AftCone
// 24 RBlivet
// 25 LBlivet
// 26 AftHull2
// 27 RInrDr
// 28 ROutrDr
// 29 LInrDr
// 30 LOutrDr
// 31 Pitot
// 32 TailBull
// 33 Gun

/*
	static UINT LPLDoorGrp[1] = {0};
	static MGROUP_ROTATE LPLDoor (0, LPLDoorGrp, 1,
		_V(-0.488,0.774,0.0), _V(0,0,1), (float)( 150*RAD));
	static UINT RPLDoorGrp[1] = {0};
	static MGROUP_ROTATE RPLDoor (0, RPLDoorGrp, 1,
		_V( 0.488,0.774,0.0), _V(0,0,1), (float)(-150*RAD));
	static UINT HatchGrp[1] = {0};
	static MGROUP_ROTATE Hatch (0, HatchGrp, 1,
		_V( 0.0,1.049,1.978), _V(1,0,0), (float)( 155*RAD));
//		_V(-0.422,0.8,6.25), _V(0,0,1), (float)( 130*RAD));
*/

	AddAnimationComponent (anim_gear, 0.05, 0.36, &LNoseGrDr);
	AddAnimationComponent (anim_gear, 0.05, 0.36, &RNoseGrDr);
	AddAnimationComponent (anim_gear, 0.37, 0.91, &NoseGearArm);
	AddAnimationComponent (anim_gear, 0.37, 0.91, &NoseGearYoke);
	AddAnimationComponent (anim_gear, 0.37, 0.91, &NoseGearStrut);


	AddAnimationComponent (anim_gear, 0.05, 0.4, &RInrGrDr);
	AddAnimationComponent (anim_gear, 0.05, 0.4, &ROutrGrDr);
	AddAnimationComponent (anim_gear, 0.41, 0.95, &RLongStrut);
	AddAnimationComponent (anim_gear, 0.41, 0.95, &RAFrame);
	AddAnimationComponent (anim_gear, 0.41, 0.95, &RStrut);
	AddAnimationComponent (anim_gear, 0.41, 0.95, &RWheel);


	AddAnimationComponent (anim_gear, 0.05, 0.41, &LInrGrDr);
	AddAnimationComponent (anim_gear, 0.05, 0.41, &LOutrGrDr);
	AddAnimationComponent (anim_gear, 0.42, 0.97, &LLongStrut);
	AddAnimationComponent (anim_gear, 0.42, 0.97, &LAFrame);
	AddAnimationComponent (anim_gear, 0.42, 0.97, &LStrut);
	AddAnimationComponent (anim_gear, 0.42, 0.97, &LWheel);

/*	Gear1 = AddAnimationComponent (anim_gear, 0.4, 0.75, &LFSupt);
	AddAnimationComponent (anim_gear, 0.75, 0.95, &LFGear, Gear1);
	Gear2 = AddAnimationComponent (anim_gear, 0.4, 0.75, &LASupt);
	AddAnimationComponent (anim_gear, 0.75, 0.95, &LAGear, Gear2);
	Gear3 = AddAnimationComponent (anim_gear, 0.4, 0.75, &RFSupt);
	AddAnimationComponent (anim_gear, 0.75, 0.95, &RFGear, Gear3);
	Gear4 = AddAnimationComponent (anim_gear, 0.4, 0.75, &RASupt);
	AddAnimationComponent (anim_gear, 0.75, 0.95, &RAGear, Gear4);
	AddAnimationComponent (anim_gear, 0.05, 0.95, &NoseGear);
*/



	// ***** Elevon animation *****
	static UINT LElevonGrp[1] = {1};
	static MGROUP_ROTATE LElevon (0, LElevonGrp, 1,
		_V(-4.1,-0.28,-8.094), _V(-1,0.0, 0), (float)(-25*RAD));
	AddAnimationComponent (anim_lelevon, 0, 1, &LElevon);
	static UINT RElevonGrp[1] = {2};
	static MGROUP_ROTATE RElevon (0, RElevonGrp, 1,
		_V( 4.1,-0.28,-8.094), _V( 1,0.0, 0), (float)( 25*RAD));
	AddAnimationComponent (anim_relevon, 0, 1, &RElevon);
/*
	// ***** Elevator animation *****
	static UINT LElevatorGrp[1] = {32};
	static MGROUP_ROTATE LElevator (0, LElevatorGrp, 1,
		_V(-4.35,1.1,-30.57), _V(-1,0.0, 0.3), (float)(-25*RAD));
	AddAnimationComponent (anim_lelevator, 0, 1, &LElevator);
	static UINT RElevatorGrp[1] = {33};
	static MGROUP_ROTATE RElevator (0, RElevatorGrp, 1,
		_V( 4.35,1.1,-30.57), _V( 1,0.0, 0.3), (float)( 25*RAD));
	AddAnimationComponent (anim_relevator, 0, 1, &RElevator);
/*
	// ***** Elevator trim animation *****
	static UINT LElevatorTrimGrp[1] = {3};
	static MGROUP_ROTATE LElevatorTrim (5, LElevatorTrimGrp, 1,
		_V(-2.15,0.83,-3.74), _V(-0.25,0.45,-0.4), (float)(-10*RAD));
	anim_lelevatortrim = CreateAnimation (0.5);
	AddAnimationComponent (anim_lelevatortrim, 0, 1, &LElevatorTrim, wing1);
	static UINT RElevatorTrimGrp[1] = {5};
	static MGROUP_ROTATE RElevatorTrim (5, RElevatorTrimGrp, 1,
		_V( 2.15,0.83,-3.74), _V( 0.25,0.45,-0.4), (float)( 10*RAD));
	anim_relevatortrim = CreateAnimation (0.5);
	AddAnimationComponent (anim_relevatortrim, 0, 1, &RElevatorTrim, wing2);
*/
/*	// ***** Aileron animation *****
	static UINT LAileronGrp[1] = {36};
	static MGROUP_ROTATE LAileron (0, LAileronGrp, 1,
		_V( 27.95,2.17,-20.27), _V(-1,0.1, 0.15), (float)(-20*RAD));
	AddAnimationComponent (anim_laileron, 0, 1, &LAileron, wing1);

	static UINT RAileronGrp[1] = {35};
	static MGROUP_ROTATE RAileron (0, RAileronGrp, 1,
		_V(-27.95,2.17,-20.27), _V( 1,0.1, 0.15), (float)(-20*RAD));
	AddAnimationComponent (anim_raileron, 0, 1, &RAileron, wing2);
*/
	// ***** Rudder animation *****
	static UINT RudderGrp[1] = {3};
	static MGROUP_ROTATE Rudder (0, RudderGrp, 1,
		_V( 0,3.625,-12.156), _V(0,0.85,-0.55), (float)(-20*RAD));
	AddAnimationComponent (anim_rudder, 0, 1, &Rudder);



}


void B58::MakeAirfoils()
{
	ClearAirfoilDefinitions();
	ClearVariableDragElements();


		CreateVariableDragElement (&gear_proc, 0.8, _V(0, -1, 0));


//					align,         attack pt,       coeff func, c, S,   AR
	CreateAirfoil (LIFT_VERTICAL, _V(0,0,-0.1),    VLiftCoeff, 12, 143, 6.55);
	// wing and body lift+drag components

	CreateAirfoil (LIFT_HORIZONTAL, _V(0,0,-12.5), HLiftCoeff, 4,  15, 3.625);
	// vertical stabiliser and body lift and drag components


	MakeControlSurfaces();
}


void B58::MakeControlSurfaces()
{

//  Aargh!  Can't delete control surfaces
//  Need to make sure we declare once and only once

	if (bControlsMade) return;
	bControlsMade = true;

	CreateControlSurface (AIRCTRL_ELEVATOR,      3.5,  0.7, _V(  3.2,0,-8.0), AIRCTRL_AXIS_XPOS, anim_lelevator);
	CreateControlSurface (AIRCTRL_ELEVATOR,      3.5,  0.7, _V( -3.2,0,-8.0), AIRCTRL_AXIS_XPOS, anim_relevator);
	CreateControlSurface (AIRCTRL_RUDDER,        3.0,  0.7, _V(   0, 0,-12.0), AIRCTRL_AXIS_YPOS, anim_rudder);
	CreateControlSurface (AIRCTRL_AILERON,       8.1, -0.9, _V(-4.2,0,-8.0), AIRCTRL_AXIS_XPOS, anim_raileron);
	CreateControlSurface (AIRCTRL_AILERON,       8.1, -0.9, _V( 4.2,0,-8.0), AIRCTRL_AXIS_XNEG, anim_laileron);
	CreateControlSurface (AIRCTRL_ELEVATORTRIM,  0.6,  0.7, _V(  3.2,0,-8.0), AIRCTRL_AXIS_XPOS);
	CreateControlSurface (AIRCTRL_ELEVATORTRIM,  0.6,  0.7, _V( -3.2,0,-8.0), AIRCTRL_AXIS_XPOS);

}




void B58::UpdateAnimations()
{	

	// animate landing gear
	if (gear_status >= DOOR_CLOSING) {
		double da = oapiGetSimStep() * GEAR_OPERATING_SPEED;
		if (gear_status == DOOR_CLOSING) {
			if (gear_proc > 0.0)
				gear_proc = max (0.0, gear_proc-da);
			else {
				gear_status = DOOR_CLOSED;
			}
		} else  { // door opening
			if (gear_proc < 1.0)
				gear_proc = min (1.0, gear_proc+da);
			else {
				gear_status = DOOR_OPEN;
			}
		}
		SetAnimation (anim_gear, gear_proc);
	}

	lelevon_des = 0.5+(GetManualControlLevel(THGROUP_ATT_PITCHUP, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE) - 
			GetManualControlLevel(THGROUP_ATT_PITCHDOWN, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE)+GetManualControlLevel(THGROUP_ATT_BANKLEFT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE) +
			0.3*GetControlSurfaceLevel (AIRCTRL_ELEVATORTRIM) -			
			GetManualControlLevel(THGROUP_ATT_BANKRIGHT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE));
	relevon_des = 0.5+(GetManualControlLevel(THGROUP_ATT_PITCHUP, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE) - 
			GetManualControlLevel(THGROUP_ATT_PITCHDOWN, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE)+GetManualControlLevel(THGROUP_ATT_BANKRIGHT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE) +
			0.3*GetControlSurfaceLevel (AIRCTRL_ELEVATORTRIM) -			 
			GetManualControlLevel(THGROUP_ATT_BANKLEFT, MANCTRL_ANYDEVICE, MANCTRL_ANYMODE));

	if (relevon_des > relevon_proc)
		relevon_proc += 0.03;
	if (relevon_des < relevon_proc)
		relevon_proc -= 0.03;
	if (lelevon_des > lelevon_proc)
		lelevon_proc += 0.03;
	if (lelevon_des < lelevon_proc)
		lelevon_proc -= 0.03;
	SetAnimation (anim_lelevon, lelevon_proc);
	SetAnimation (anim_relevon, relevon_proc);

}

void B58::ActivateLandingGear (DoorStatus action)
{
	gear_status = action;
}

void B58::RevertLandingGear ()
{
	ActivateLandingGear (gear_status == DOOR_CLOSED || gear_status == DOOR_CLOSING ?
						 DOOR_OPENING : DOOR_CLOSING);
}


void B58::AirBreathingEngine()
// Thanks to David Hopkins for publishing his AirBreathingEngine source.  
{//=========================================================
double alt2;
double eas2;
double cf1, cf2;
double ActualThrust;


alt2 = GetAltitude();
eas2 = GetAirspeed ();

alt2 = alt2*3.28;
//alt2 = alt2*0.78;		// a kludge that keeps the "airbreathing" engine operating at higher altitude
eas2 = eas2*2.236;

cf1 = 1.67e-6*(eas2*eas2)-8.8789e-4*(eas2)+0.9977;	
cf2 = 3.0e-10*(alt2*alt2)-3.0000e-5*(alt2)+0.9934;

// Nasty kludge to keep the thrust from increasing dramatically above 100kft
if (alt2 > 50000)
	cf2 = 0.25*(1-((alt2-50000)*.0001));
if (cf2 < 0) cf2 = 0;
if (cf1 > 3.0) cf1 = 3.0;

ActualThrust = cf1*cf2*THRUST_JET_VAC;
if (GetMachNumber() > 1.9)
	ActualThrust = ActualThrust * (1-(GetMachNumber()*2-1.9*2));

//sprintf(oapiDebugString(), "cf1: %.3f, cf2:  %.3f, cf1*cf2:  %.3f ActualThrust:  %.3f", cf1, cf2, cf1*cf2, ActualThrust);

SetThrusterMax0 (th_first[0],ActualThrust);
SetThrusterMax0 (th_first[1],ActualThrust);
SetThrusterMax0 (th_first[2],ActualThrust);
SetThrusterMax0 (th_first[3],ActualThrust);

SetThrusterMax0 (th_AB[0],ActualThrust);
SetThrusterMax0 (th_AB[1],ActualThrust);
SetThrusterMax0 (th_AB[2],ActualThrust);
SetThrusterMax0 (th_AB[3],ActualThrust);

double ThrusterActive = GetThrusterGroupLevel(THGROUP_MAIN);


		if((ThrusterActive>0.8) && (cf1 * cf2 > 0))
			{
			// we calculate the volume at wich it must be played
			double volume=110+(145*ThrusterActive);
			double freq=11050+(9000*ThrusterActive);
			PlayVesselWave3(MyID,PLAYJET,LOOP,volume,freq); // play the sound
			SoundOptionOnOff3(MyID,PLAYMAINTHRUST,true);								// Rocket noise too
			SetThrusterGroupLevel(thg_AB,ThrusterActive);
//			SetThrusterGroupLevel(thg_AB,1.0);
			}
		else if((ThrusterActive>0.0) && (cf1 * cf2 > 0))
			{
			// we calculate the volume at wich it must be played
			double volume=110+(145*ThrusterActive);
			double freq=11050+(9000*ThrusterActive);
			PlayVesselWave3(MyID,PLAYJET,LOOP,volume,freq); // play the sound
			SoundOptionOnOff3(MyID,PLAYMAINTHRUST,false);								// No rocket noise
			SetThrusterGroupLevel(thg_AB,0.0);
			}
		else if((IsPlaying3(MyID,PLAYJET)) || (cf1 * cf2 < 0.01))
			{
			// else shut down the sound if it's playing
			StopVesselWave3(MyID,PLAYJET);												// No jet noise
			SoundOptionOnOff3(MyID,PLAYMAINTHRUST,false);								// No rocket noise
			SetThrusterGroupLevel(thg_AB,0.0);
			}


//SetThrusterMax0 (th_first[0],ActualThrust);
};
	
void B58::SetGearParameters (double state)
{

	double geary = ((1.0-state)*1.0);
	if ((gear_status == DOOR_OPENING) || (gear_status == DOOR_CLOSING)) {
		SetTouchdownPoints (_V(0,-2.77+0.6*geary,9.78), _V(-2.17,-2.82+0.6*geary,-2.34), _V( 2.17,-2.82+0.6*geary,-2.34));
	}
	if (gear_proc == 1.0) {
		SetSurfaceFrictionCoeff (0.025, 0.4);
	} else 
	{
		SetSurfaceFrictionCoeff (5, 0.5);		
		if(GroundContact()) 
		{
			//crashed on the ground

			gear_proc = 0.0;
			gear_status = DOOR_OPEN;
		}
		
	}


	double steerforce;
	double steerangle;

	steerforce = 40.0;
	if (GetAirspeed() < 40.0)
		steerforce *=8.0;
	steerforce = steerforce*GetMass();
	if (GetAirspeed () < 5.0) steerforce*=(GetAirspeed()/5); // from 1* to 0*

	steerangle = GetControlSurfaceLevel (AIRCTRL_RUDDER);
	steerforce = steerforce * steerangle;
	if ((GetAltitude() > 5.5) || (state < 1.0)) steerforce = 0.0;

	AddForce (_V(steerforce, 0, 0), _V(0, 0, 12.0));
	AddForce (_V(-steerforce, 0, 0), _V(0, 0, -12.0));


	// sprintf (oapiDebugString(), "%0.4f", steerforce );


}


void B58::TimeStep(  double simt )
{
	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* This function loops through the pending GPDS requests and returns a bool (true/false)
	/* if any GPDS requests have been implemented.
	/* If any have (eg adding a new payload), you will need to re-draw your ship, hence the
	/* call to the (repeatable) SetMainStage(); */

	if( CheckCentralMessage() ) {
		while( CheckCentralMessage() ) ;
			SetMainStage();
	}


	if (!bControlsMade)
		MakeControlSurfaces();



	AirBreathingEngine();
/*	double ThrusterActive = GetThrusterGroupLevel(THGROUP_MAIN);
	if(ThrusterActive>0.0f)
		{
		// we calculate the volume at wich it must be played
		double volume=110+(145*ThrusterActive);
		double freq=11050+(9000*ThrusterActive);
		PlayVesselWave3(MyID,PLAYJET,LOOP,volume,freq); // play the sound
		}
	else if(IsPlaying3(MyID,PLAYJET))
		{
		// else shut down the sound if it's playing
		StopVesselWave3(MyID,PLAYJET);
		}
*/

	UpdateAnimations();
	SetGearParameters(gear_proc);

	if (stage==0)
		if (ph_tank)
			if (GetPropellantMass(ph_tank) < 0.1)
				JettisonPart();
}


/* CVE-lite function ! */
/* IMPORTANT */
/* VisualCreated() is a function passed through from CVE-lite as a direct replacement
/* for the DLLCLBK ovcVisualCreated() function.
/* You can't use ovcVisualCreated() with CVE-lite, so use this instead! */


void B58::VisualCreated ( VESSEL *vessel, VISHANDLE vis, int refcount) {
	/* local equivalent to the DLLCLBK ovcVisualCreated() */
	SetMainStage(); // this is a good place to ask for the initial setup.
}


void B58::SetClassCaps(  FILEHANDLE cfg )
{

	CheckCentralMessage();

	SetMainStage();
}

void B58::SaveState( FILEHANDLE scn )
{
	char cbuf[256];

	SaveDefaultState (scn);
	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Call this after the normal SaveDefautState() to record attached payloads
	/* it the scenario file */


//	Undock(ALLDOCKS);

	oapiWriteScenario_int (scn, "CONFIGURATION", stage);
	sprintf (cbuf, "%d %0.4f", gear_status, gear_proc);
	oapiWriteScenario_string (scn, "GEAR", cbuf);
	sprintf (cbuf, meshname);
	oapiWriteScenario_string (scn, "MESH", cbuf);

	/* save payload lines */
	SaveDefaultStateEx(scn);
}

void B58::LoadState(  FILEHANDLE scn, void *vs )
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) 
	{
		/* CVE-lite function ! */
		/* IMPORTANT */
		/* Call this for every line in the scenario file. It will automatically
		/* deduce which lines are for it, and load the payloads from the file. */

		/* load payload lines */
		if (!strnicmp (line, "GEAR", 4)) {
		sscanf (line+4, "%d%lf", &gear_status, &gear_proc);
		} else if (!strnicmp (line, "CONFIGURATION", 13)) {
		    sscanf (line+13, "%d", &stage);
		} else if(!strnicmp (line, "MESH", 4) ) {
			sscanf( line+4, "%s", &meshname);

		} else ParsePayloadString(line);

		ParseScenarioLineEx (line, vs);
	}



	/* always set the stage whenever something may have changed */
	SetMainStage();

	RevertLandingGear();
	RevertLandingGear();

}


void B58::PlaceComponents()
{
	COMMON_MESH tmp;


	tmp.MeshHandle(hB58);
	tmp.MeshPosition(_V(0,0,0));
	AddMeshEx (tmp);
	if (stage < 2) {
		tmp.MeshHandle(hWeapon);
		tmp.MeshPosition(_V(0,-1.5,4.0));
		AddMeshEx (tmp);
	}
	if (stage < 1) {
		tmp.MeshHandle(hTank);
		tmp.MeshPosition(_V(0,-1.5,4.0));
		AddMeshEx (tmp);
	}

}

/* IMPORTANT */
/* Always have a function that you can call as many times as you like that will re-draw
/* all you meshes, reset all your thrusters and check to make sure the propellent is
/* there. This should be called after any configuration change, or any GDPS line (see
/* TimeStep(), above) */

void B58::SetMainStage( )
{

	SoundOptionOnOff3(MyID,PLAYMAINTHRUST,true);

	SetB58Stage();
}


void B58::SetB58Stage( )
{


	/* CVE-lite function ! */
	/* ClearMeshesEx() will get rid of all the normal meshes, plus anything you've defined
	/* with the COMMON_MESH class (not used here) */

	/* erase all meshes and thrusters so we can start afresh */
	ClearMeshesEx();
	ClearThrusterDefinitions();

	PARTICLESTREAMSPEC contrail = 
	{
		0, 1.1, 30, 25.0, 0, 0.5, -.5, 0.5, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 0.1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC exhaust_main = {
		0, 3, 25, 350, 0.1, 0.2, 28, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	};
	exhaust_main.tex = oapiRegisterParticleTexture ("Contrail2");
	SetSize (30);

	/* CVE-lite function ! */
	/* IMPORTANT */
	/* double CalculateMass() will return the all-up mass of the combined payloads and
	/* upper stages. Simply add this to your normal dry mass. */
	
	/* always add payload mass [CalculateMass()] to the vessel's empty mass */
	double addmass = 0.0;
	if (stage == 1) addmass += bombmass;
	if (stage == 0) addmass += tankmass;
	SetEmptyMass (JET_DRYMASS + addmass + CalculateMass() );

	SetTouchdownPoints (_V(0,-2.77,9.78), _V(-2.17,-2.82,-2.34), _V( 2.17,-2.82,-2.34));
	SetSurfaceFrictionCoeff (0.025, 0.4);
//	SetTouchdownPoints (_V(0,-10,-11), _V(-10,10,-11), _V(10,10,-11));
	SetPMI (_V(75,42.6,13.8));
	SetCrossSections (_V(16.24,63.8,9.0));
	SetCW (0.01, 0.2, 0.7, 0.7);
	SetRotDrag (_V(0.7,0.7,1.2));
	SetPitchMomentScale (1e-6);
	SetBankMomentScale (1e-6);
//	SetLiftCoeffFunc (0); 
	SetTrimScale (0.025);
	SetCameraOffset(CAMERA_OFS);
	MakeAirfoils();
//	SetControlSurfaceLevel(AIRCTRL_ELEVATORTRIM,1.0);
	
	/* CVE-lite function ! */
	/* IMPORTANT */
	/* Draw() should be called to render all the loaded payloads, plus anything you've
	/* added with the COMMON_MESH class (not used here) */

	/* call draw to render the payloads, plus anything you've added from COMMON_MESH */
	PlaceComponents();
	Draw();
	
	SetMaxWheelbrakeForce (7.5*GetMass());

	/* CVE-lite variable ! */
	/* IMPORTANT */
	/* Use the keyword double InitFuelLevel multipled by the maximum mass to have the
	/* fuel levels automatically set from the GPDS payload string */

	/* verify that the propellent hasn't been created before making it */
	if (!ph_b58)  
		ph_b58  = CreatePropellantResource(JET_PROP_MASS,JET_PROP_MASS,1.0); 

	if (stage > 0) {

		SetDefaultPropellantResource (ph_b58); 

		th_first[0] = CreateThruster ( _V( 6.68, 0.0,-8.21), _V(0,0,1), THRUST_JET_VAC, ph_b58, ISP_JET_VAC, ISP_JET_SL);
	//	AddExhaustStream (th_first[0], _V( 6.68,-1.24,-8.21), &exhaust_main);
		th_first[1] = CreateThruster ( _V( 3.86, 0.0,-1.46), _V(0,0,1), THRUST_JET_VAC, ph_b58, ISP_JET_VAC, ISP_JET_SL);
	//	AddExhaustStream (th_first[1], _V( 3.86,-1.52,-1.46), &exhaust_main);
		th_first[2] = CreateThruster ( _V(-3.86, 0.0,-1.46), _V(0,0,1), THRUST_JET_VAC, ph_b58, ISP_JET_VAC, ISP_JET_SL);
	//	AddExhaustStream (th_first[2], _V(-3.86,-1.52,-1.46), &exhaust_main);
		th_first[3] = CreateThruster ( _V(-6.68, 0.0,-8.21), _V(0,0,1), THRUST_JET_VAC, ph_b58, ISP_JET_VAC, ISP_JET_SL);
	//	AddExhaustStream (th_first[3], _V(-6.68,-1.24,-8.21), &exhaust_main);


		th_AB[0] = CreateThruster ( _V( 6.68, 0.0,-8.21), _V(0,0,1), THRUST_JET_VAC, ph_b58, ISP_JET_VAC, ISP_JET_SL);
		AddExhaustStream (th_AB[0], _V( 6.68,-1.24,-8.21), &contrail);
		th_AB[1] = CreateThruster ( _V( 3.86, 0.0,-1.46), _V(0,0,1), THRUST_JET_VAC, ph_b58, ISP_JET_VAC, ISP_JET_SL);
		AddExhaustStream (th_AB[1], _V( 3.86,-1.52,-1.46), &contrail);
		th_AB[2] = CreateThruster ( _V(-3.86, 0.0,-1.46), _V(0,0,1), THRUST_JET_VAC, ph_b58, ISP_JET_VAC, ISP_JET_SL);
		AddExhaustStream (th_AB[2], _V(-3.86,-1.52,-1.46), &contrail);
		th_AB[3] = CreateThruster ( _V(-6.68, 0.0,-8.21), _V(0,0,1), THRUST_JET_VAC, ph_b58, ISP_JET_VAC, ISP_JET_SL);
		AddExhaustStream (th_AB[3], _V(-6.68,-1.24,-8.21), &contrail);
	}
	else {

		if (!ph_tank)  
			ph_tank  = CreatePropellantResource(TANK_PROP_MASS,TANK_PROP_MASS,1.0); 
		SetDefaultPropellantResource (ph_tank); 
	
		th_first[0] = CreateThruster ( _V( 6.68, 0.0,-8.21), _V(0,0,1), THRUST_JET_VAC, ph_tank, ISP_JET_VAC, ISP_JET_SL);
	//	AddExhaustStream (th_first[0], _V( 6.68,-1.24,-8.21), &exhaust_main);
		th_first[1] = CreateThruster ( _V( 3.86, 0.0,-1.46), _V(0,0,1), THRUST_JET_VAC, ph_tank, ISP_JET_VAC, ISP_JET_SL);
	//	AddExhaustStream (th_first[1], _V( 3.86,-1.52,-1.46), &exhaust_main);
		th_first[2] = CreateThruster ( _V(-3.86, 0.0,-1.46), _V(0,0,1), THRUST_JET_VAC, ph_tank, ISP_JET_VAC, ISP_JET_SL);
	//	AddExhaustStream (th_first[2], _V(-3.86,-1.52,-1.46), &exhaust_main);
		th_first[3] = CreateThruster ( _V(-6.68, 0.0,-8.21), _V(0,0,1), THRUST_JET_VAC, ph_tank, ISP_JET_VAC, ISP_JET_SL);
	//	AddExhaustStream (th_first[3], _V(-6.68,-1.24,-8.21), &exhaust_main);


		th_AB[0] = CreateThruster ( _V( 6.68, 0.0,-8.21), _V(0,0,1), THRUST_JET_VAC, ph_tank, ISP_JET_VAC, ISP_JET_SL);
		AddExhaustStream (th_AB[0], _V( 6.68,-1.24,-8.21), &contrail);
		th_AB[1] = CreateThruster ( _V( 3.86, 0.0,-1.46), _V(0,0,1), THRUST_JET_VAC, ph_tank, ISP_JET_VAC, ISP_JET_SL);
		AddExhaustStream (th_AB[1], _V( 3.86,-1.52,-1.46), &contrail);
		th_AB[2] = CreateThruster ( _V(-3.86, 0.0,-1.46), _V(0,0,1), THRUST_JET_VAC, ph_tank, ISP_JET_VAC, ISP_JET_SL);
		AddExhaustStream (th_AB[2], _V(-3.86,-1.52,-1.46), &contrail);
		th_AB[3] = CreateThruster ( _V(-6.68, 0.0,-8.21), _V(0,0,1), THRUST_JET_VAC, ph_tank, ISP_JET_VAC, ISP_JET_SL);
		AddExhaustStream (th_AB[3], _V(-6.68,-1.24,-8.21), &contrail);
	}

	thg_first = CreateThrusterGroup (th_first, 4, THGROUP_MAIN);
	thg_AB = CreateThrusterGroup (th_AB, 4, THGROUP_USER);

	AddExhaust (th_AB[0], 15.0, 1.5, _V( 6.68,-1.24,-8.21), _V(0,0,-1), tex);
	AddExhaust (th_AB[1], 15.0, 1.5, _V( 3.86,-1.52,-1.46), _V(0,0,-1), tex);
	AddExhaust (th_AB[2], 15.0, 1.5, _V(-3.86,-1.52,-1.46), _V(0,0,-1), tex);
	AddExhaust (th_AB[3], 15.0, 1.5, _V(-6.68,-1.24,-8.21), _V(0,0,-1), tex);
//	numengines = 4;

	SoundOptionOnOff3(MyID,PLAYMAINTHRUST,false);

	EnableTransponder(true);

}

void B58::JettisonPart()
 {

	 
				double throttle = GetThrusterGroupLevel(THGROUP_MAIN);
				VESSELSTATUS vs;
				VECTOR3 ofs = _V(0,-1.5,4.0);
				VECTOR3 vel = _V(0,0,0);
				GetStatus (vs);
				VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
				Local2Rel (ofs, vs.rpos);
				GlobalRot (vel, rofs);
				vs.rvel.x = rvel.x+rofs.x;
				vs.rvel.y = rvel.y+rofs.y;
				vs.rvel.z = rvel.z+rofs.z;
				vs.vrot.x = 0.0;
				vs.vrot.y = 0.0;
				vs.vrot.z = 0.0;
				vs.fuel = 0.0;
				char name[256];
				if (stage == 0) {
					strcpy (name, GetName()); strcat (name, "-Tank");
					VESSEL::Create (name, "BLU2p1", vs);
					DelPropellantResource (ph_tank); 	
				} else if (stage == 1) {
					strcpy (name, GetName()); strcat (name, "-Weapon");
					VESSEL::Create (name, "BLU2p2", vs);	
				}
				PlayVesselWave3(MyID,PLAYBANG);

				stage++;
				SetMainStage();
				SetThrusterGroupLevel(THGROUP_MAIN,throttle);
 }

int B58::ConsumeBufferedKey(DWORD dwKey, bool fPressed, const char *keystate)
{
	if (!fPressed) return 0; // only process keydown events
	if (KEYMOD_SHIFT (keystate)) 
	{
	}
	else if (KEYMOD_CONTROL (keystate)) 
	{
		// insert ctrl key combinations here
	}
	else { // unmodified keys
		if (KEYDOWN (keystate, OAPI_KEY_J)) { // "Jettison Stage"
			if (oapiAcceptDelayedKey (OAPI_KEY_J, 1.0)) {
				if (stage == 2) {
					VESSEL *depV;
					OBJHANDLE hVessel;
					double throttle = GetThrusterGroupLevel(THGROUP_MAIN);
//					OverridePayloadVelocity(_V(0,-5.0,-1.0));
					depV = DeployPayload();

					VESSELSTATUS vs, vsPL;
					
					VECTOR3 ofs = _V(0,0,0);
					VECTOR3 vel = _V(0,0,0);
					GetStatus (vs);
					depV->GetStatus(vsPL);
					VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
					Local2Rel (ofs, vs.rpos);
					GlobalRot (vel, rofs);
					vs.rvel.x = rvel.x+rofs.x;
					vs.rvel.y = rvel.y+rofs.y;
					vs.rvel.z = rvel.z+rofs.z;
					vsPL.rvel = vs.rvel;
					depV->DefSetState(&vsPL);

	//				hVessel=oapiGetVesselByName(depV->GetName());
	//				if (hVessel != 0)
	//					oapiSetFocusObject(hVessel);
						

						/* CVE-lite function ! */
						/* IMPORTANT */
						/* call DeployPayload() to jettison the GPDS payloads in sequence */

						/* IMPORTANT */
						/* Always re-set the stage after any kind of change */
					SetMainStage();
					SetThrusterGroupLevel(THGROUP_MAIN,throttle);
				}
				else JettisonPart();
			}
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_B)) 
		{ // Toggle the brakes
			if (oapiAcceptDelayedKey (OAPI_KEY_B, 1.0))
			{
				if (BrakesOn)
					BrakesOn = false;
				else BrakesOn = true;
			}
			return 1;
		}
		if (KEYDOWN (keystate, OAPI_KEY_G)) {
			if (oapiAcceptDelayedKey (OAPI_KEY_G, 1.0)) {
				// Toggle the Gear
				RevertLandingGear ();
				StopVesselWave3(MyID,PLAYGEAR);
				PlayVesselWave3(MyID,PLAYGEAR);
			}
			return 1;
		}

	}
	return 0;
}



int B58::ConsumeKey( const char *keystate )
{

	if (KEYMOD_SHIFT (keystate)) 
	{
						
		return 0; // shift combinations are reserved
	}
	else if (KEYMOD_CONTROL (keystate)) 
	{
		return 0;
	}
	else 
	{ // unmodified keys
	
		return 0;

	}
	return 0;
}



